@@ -6,7 +6,7 @@
 #
 
 package File::ByLine;
-$File::ByLine::VERSION = '1.181950'; # TRIAL
+$File::ByLine::VERSION = '1.181951'; # TRIAL
 use v5.10;
 
 # ABSTRACT: Line-by-line file access loops
@@ -162,7 +162,7 @@ File::ByLine - Line-by-line file access loops
 
 =head1 VERSION
 
-version 1.181950
+version 1.181951
 
 =head1 SYNOPSIS
 
@@ -260,8 +260,8 @@ needed.
 This function calls a coderef once for each line in the file.  The file is read
 line-by-line, removes the newline character(s), and then executes the coderef.
 
-Each line (without newline) is passed to the coderef as the first parameter and
-only parameter to the coderef.  It is also placed into C<$_>.
+Each line (without newline) is passed to the coderef as the only parameter to
+the coderef.  It is also placed into C<$_>.
 
 This function returns the number of lines in the file.
 
@@ -283,8 +283,8 @@ file.
 This function calls a coderef once for each line in the file.  The file is read
 line-by-line, removes the newline character(s), and then executes the coderef.
 
-Each line (without newline) is passed to the coderef as the first parameter and
-only parameter to the coderef.  It is also placed into C<$_>.
+Each line (without newline) is passed to the coderef as the only parameter to
+the coderef.  It is also placed into C<$_>.
 
 This function returns the number of lines in the file.
 
@@ -375,8 +375,8 @@ the return value of that coderef, returns only the lines where the coderef
 evaluates to true.  This is similar to the C<grep> built-in function, except
 operating on file input rather than array input.
 
-Each line (without newline) is passed to the coderef as the first parameter and
-only parameter to the coderef.  It is also placed into C<$_>.
+Each line (without newline) is passed to the coderef as the only parameter to
+the coderef.  It is also placed into C<$_>.
 
 This function returns the lines for which the coderef evaluates as true.
 
@@ -428,8 +428,8 @@ basically if the coderef returns a list, all elements of that list are added
 as distinct elements to the return value array.  If the coderef returns an
 empty list, no elements are added.
 
-Each line (without newline) is passed to the coderef as the first parameter and
-only parameter to the coderef.  It is also placed into C<$_>.
+Each line (without newline) is passed to the coderef as the only parameter to
+the coderef.  It is also placed into C<$_>.
 
 This is meant to be similar to the built-in C<map> function.
 
@@ -492,29 +492,62 @@ Constructs a new object, suitable for the object oriented calls below.
 
 =head2 ATTRIBUTES
 
+=head3 extended_info
+
+  $byline->extended_info(1);
+
+This was added in 1.181951.
+
+Gets and sets the "extended information" flag.  This defaults to false, but
+if set to a true value this will pass a second parameter to all user-defined
+code (such as the per-line code function in C<dolines> and C<do> and the
+C<header_handler> function.
+
+For all code, this information will be passed as the second argument to the
+user defined code.  It will be a hashref with the following keys defined:
+
+=over 4
+
+=item C<filename> - The filename currently being processed
+
+=item C<object> - An object corresponding to either the current explicit or implicit C<File::ByLine> object
+
+=item C<process_number> - Which child process (first process is zero)
+
+=back
+
+This object should not be modified by user code.  In addition, no attributes
+of the explict or implicit File::ByLine object passed as part of this hashref
+should be modified within user code.
+
 =head3 file
 
   my $current_file = $byline->file();
   $byline->file("abc.txt");
+  $byline->file( [ "abc.txt", "def.txt" ] );
+  $byline->file( "$abc.txt", "def.txt" );
 
 Gets and sets the default filename used by the methods in the object oriented
 interface.  The default value is C<undef> which indicates that no default
 filename is provided.
 
-Instead of a single filename, an arrayref can be passed in, in which case the
-files are read in turn as if they are all one file. Note that if the file
-doesn't end in a newline, a newline is inserted before processing the next
+Instead of a single filename, a list or arrayref can be passed in, in which
+case the files are read in turn as if they are all one file. Note that if the
+file doesn't end in a newline, a newline is inserted before processing the next
 file.
 
-=head3 header_skip
+=head3 header_all_files
 
-  $byline->header_skip(1);
+  $byline->header_all_files(1);
 
-Gets and sets whether the object oriented methods will skip the first line
-in the file (which you might want to do for a line that is a header).  This
-defaults to false.  Any true value will cause the header line to be skipped.
+Gets and sets whether the object oriented methods will call C<header_handler>
+for every file if multiple files are passed into the C<file> attribute.
 
-You cannot set this to true while a C<header_handler> value is set.
+The anticipated usage of this would be with C<extended_info> set to true, with
+the C<header_handler> function examining the C<filename> attribute of the
+extended info hashref.  Note that all headers may be read before any line in
+any file is read, to better accommodate parallel code execution.  I.E. the
+headers of all files may be read at once before any data line is read.
 
 =head3 header_handler
 
@@ -526,8 +559,9 @@ When a header handler is specified, the first row of the file is sent to this
 handler, and is not sent to the code provided to the various do/grep/map/lines
 methods in the object oriented interface.
 
-The code is called with one parameter, the header line.  The header line is
-also stored in C<$_>.
+The code is called with one or two parameters, the header line, and, if the
+C<extended_info> attribute is set, the extended information hashref.  The
+header line is also stored in C<$_>.
 
 When set, this is always executed in the parent process, not in the child
 processes that are spawned (in the case of C<processes> being greater than
@@ -557,8 +591,10 @@ the filename is not provided, the C<file> attribute is used for this.  See the
 C<dolines> and C<parallel_dolines> functions for more information on how this
 functions.
 
-The code is called with one parameter, the header line.  The header line is
-also stored in C<$_>.
+Each line (without newline) is passed to the coderef as the first parameter to
+the coderef.  It is also placed into C<$_>.  If the C<extended_info> attribute
+is true, the extended information hashref will be passed as the second
+parameter.
 
 Instead of a single filename, an arrayref can be passed in, in which case the
 files are read in turn as if they are all one file. Note that if the file
@@ -574,8 +610,10 @@ the filename is not provided, the C<file> attribute is used for this.  See the
 C<greplines> and C<parallel_greplines> functions for more information on how
 this functions.
 
-The code is called with one parameter, the header line.  The header line is
-also stored in C<$_>.
+Each line (without newline) is passed to the coderef as the first parameter to
+the coderef.  It is also placed into C<$_>.  If the C<extended_info> attribute
+is true, the extended information hashref will be passed as the second
+parameter.
 
 The output is a list of all input lines where the code reference produces a
 true result.
@@ -594,8 +632,10 @@ the filename is not provided, the C<file> attribute is used for this.  See the
 C<maplines> and C<parallel_maplines> functions for more information on how
 this functions.
 
-The code is called with one parameter, the header line.  The header line is
-also stored in C<$_>.
+Each line (without newline) is passed to the coderef as the first parameter to
+the coderef.  It is also placed into C<$_>.  If the C<extended_info> attribute
+is true, the extended information hashref will be passed as the second
+parameter.
 
 The output is the list produced by calling the passed-in code repeatively
 for each line of input.
