@@ -6,7 +6,7 @@
 #
 
 package File::ByLine::Object;
-$File::ByLine::Object::VERSION = '1.181950'; # TRIAL
+$File::ByLine::Object::VERSION = '1.181951'; # TRIAL
 use v5.10;
 
 # ABSTRACT: Internal object used by File::ByLine
@@ -22,19 +22,45 @@ use Scalar::Util qw(blessed reftype);
 # We do this intentionally:
 ## no critic (Subroutines::ProhibitBuiltinHomonyms)
 
+# Attributes and their accessors & defaults, used by the constructor
+# Each attribute name is the key of the hash, with the value being a
+# hashref of two values: accessor and default value.
+my (%ATTRIBUTE) = (
+    file             => [ \&file,             undef ],
+    extended_info    => [ \&extended_info,    undef ],
+    header_all_files => [ \&header_all_files, undef ],
+    header_handler   => [ \&header_handler,   undef ],
+    header_skip      => [ \&header_skip,      undef ],
+    processes        => [ \&processes,        1 ],
+);
+
 
 #
 # Attribute Accessor - file
 #
-# The file we operate on (most methods accept a file parameter - this is
-# only used if one is not set)
 sub file {
     my ($self) = shift;
     if ( scalar(@_) == 0 ) {
         return $self->{file};
     } elsif ( scalar(@_) == 1 ) {
         my $file = shift;
+        if ( !defined($file) ) { confess("Must pass a file or array ref as a file attribute") }
         return $self->{file} = $file;
+    } else {
+        return $self->{file} = [@_];
+    }
+}
+
+#
+# Attribute Accessor - extended_info
+#
+# Do we pass an extended information hash to the user process?
+sub extended_info {
+    my ($self) = shift;
+    if ( scalar(@_) == 0 ) {
+        return $self->{extended_info};
+    } elsif ( scalar(@_) == 1 ) {
+        return $self->{extended_info} = !!$_[0];    # !! to convert to fast boolean
     } else {
         confess("Invalid call");
     }
@@ -51,10 +77,16 @@ sub processes {
         return $self->{processes};
     } elsif ( scalar(@_) == 1 ) {
         my $procs = shift;
+
+        if ( !_is_number($procs) ) {
+            confess("processes only accepts integer values");
+        }
+
         if ( $procs < 1 ) {
             confess("Process count must be >= 1");
         }
         if ( $procs > 1 ) {
+            # Ensure we have the right packages installed
             $self->_require_parallel();
         }
         return $self->{processes} = $procs;
@@ -63,6 +95,21 @@ sub processes {
     }
 }
 
+#
+# Attribute Accessor - header_all_files
+#
+# If set to one, process all files for headers
+sub header_all_files {
+    my ($self) = shift;
+    if ( scalar(@_) == 0 ) {
+        return $self->{header_all_files};
+    } elsif ( scalar(@_) == 1 ) {
+        return $self->{header_all_files} = $_[0];
+    } else {
+        confess("Invalid call");
+    }
+}
+
 #
 # Attribute Accessor - header_handler
 #
@@ -111,13 +158,39 @@ sub header_skip {
 sub new {
     my $class = shift;
 
+    my %options;
+    if ( scalar(@_) == 1 ) {
+        # We assume this to be a hashref of options
+        %options = %{ $_[0] };
+    } elsif ( scalar(@_) > 1 ) {
+        if ( scalar(@_) % 2 ) {
+            confess("Must pass options in key/value form or as a hashref");
+        } else {
+            %options = (@_);
+        }
+    }
+
+    # Set defaults
     my $self = {};
-    $self->{header_handler} = undef;
-    $self->{header_skip}    = undef;
-    $self->{processes}      = 1;
+    foreach my $attr ( keys %ATTRIBUTE ) {
+        $self->{$attr} = $ATTRIBUTE{$attr}->[1];    # Default avlue
+    }
 
     bless $self, $class;
 
+    # Set attributes.  We use the accessor so we don't duplicate type
+    # checks.
+    foreach my $key ( sort keys %options ) {    # Sort for consistent tests
+        if ( exists( $ATTRIBUTE{$key} ) ) {
+            my $value = $options{$key};
+
+            # Call the accessor
+            $ATTRIBUTE{$key}->[0]( $self, $value );
+        } else {
+            confess("Invalid attribute: $key");
+        }
+    }
+
     return $self;
 }
 
@@ -135,9 +208,10 @@ sub do {
     if ( !_listlike($file) ) { $file = [$file] }
 
     if ( defined( $self->{header_handler} ) ) {
-        my $header = $_ = $self->_read_header( $file->[0] );
-        if ( defined($header) ) {
-            $self->{header_handler}($header);
+        my $fileno = 0;
+        for my $f (@$file) {
+            $self->_read_header( $f, $fileno );
+            $fileno++;
         }
     }
 
@@ -163,33 +237,11 @@ sub do {
 #
 # Finds and returns matching lines
 sub grep {
-    if ( scalar(@_) < 2 ) { confess "Invalid call"; }
+    if ( scalar(@_) < 2 ) { confess "Invalid call, too few arguments"; }
+    if ( scalar(@_) > 3 ) { confess "Invalid call, too many arguments"; }
     my ( $self, $code, $file ) = @_;
 
-    if ( !defined($file) )   { $file = $self->{file} }
-    if ( !defined($file) )   { confess "Must provide filename"; }
-    if ( !_listlike($file) ) { $file = [$file] }
-
-    if ( defined( $self->{header_handler} ) ) {
-        my $header = $_ = $self->_read_header( $file->[0] );
-        if ( defined($header) ) {
-            $self->{header_handler}($header);
-        }
-    }
-
-    my $procs = $self->{processes};
-
-    if ( $procs > 1 ) {
-        my $wu = Parallel::WorkUnit->new();
-
-        $wu->asyncs( $procs, sub { return $self->_grep_chunk( $code, $file, $procs, $_[0] ); } );
-
-        return map { @$_ } $wu->waitall();
-    } else {
-        my $lines = $self->_grep_chunk( $code, $file, 1, 0 );
-
-        return @$lines;
-    }
+    return $self->_grepmap( 'grep', $code, $file );
 }
 
 #
@@ -197,33 +249,62 @@ sub grep {
 #
 # Applies function to each entry and returns that result
 sub map {
-    if ( scalar(@_) < 2 ) { confess "Invalid call"; }
+    if ( scalar(@_) < 2 ) { confess "Invalid call, too few arguments"; }
+    if ( scalar(@_) > 3 ) { confess "Invalid call, too many arguments"; }
     my ( $self, $code, $file ) = @_;
 
+    return $self->_grepmap( 'map', $code, $file );
+}
+
+# Does the actual processing for map/grep
+sub _grepmap {
+    if ( scalar(@_) < 3 ) { confess "Invalid call, too few arguments"; }
+    if ( scalar(@_) > 4 ) { confess "Invalid call, too many arguments"; }
+    my ( $self, $type, $code, $file ) = @_;
+
     if ( !defined($file) )   { $file = $self->{file} }
     if ( !defined($file) )   { confess "Must provide filename"; }
     if ( !_listlike($file) ) { $file = [$file] }
 
     if ( defined( $self->{header_handler} ) ) {
-        my $header = $_ = $self->_read_header( $file->[0] );
-        if ( defined($header) ) {
-            $self->{header_handler}($header);
+        my $fileno = 0;
+        for my $f (@$file) {
+            $self->_read_header( $f, $fileno );
+            $fileno++;
         }
     }
 
     my $procs = $self->{processes};
 
+    # Is this a MAP or a GREP?
+    my $isgrep;
+    if ( $type eq 'grep' ) {
+        $isgrep = 1;
+    } elsif ( $type eq 'map' ) {
+        $isgrep = 0;
+    } else {
+        confess("Invalid type passed to _grepmap: $type");
+    }
+
     if ( $procs > 1 ) {
         my $wu = Parallel::WorkUnit->new();
 
-        $wu->asyncs( $procs, sub { return $self->_map_chunk( $code, $file, $procs, $_[0] ); } );
+        $wu->asyncs( $procs,
+            sub { return $self->_grepmap_chunk( $code, $file, $isgrep, $procs, $_[0] ); } );
 
-        return map { @$_ } $wu->waitall();
+        my @async_output = $wu->waitall();
+
+        my @file_output;
+        for ( my $i = 0; $i < scalar(@$file); $i++ ) {
+            push @file_output, map { $_->[$i] } @async_output;
+        }
+        return map { @$_ } @file_output;
     } else {
-        my $mapped_lines = $self->_map_chunk( $code, $file, 1, 0 );
+        my $mapped_lines = $self->_grepmap_chunk( $code, $file, $isgrep, 1, 0 );
 
-        return @$mapped_lines;
+        return map { @$_ } @$mapped_lines;
     }
+
 }
 
 #
@@ -239,8 +320,8 @@ sub lines {
     if ( !_listlike($file) ) { $file = [$file] }
 
     my @lines;
-    my $fileno = 0;
-    my $lineno = 0;
+    my $fileno        = 0;
+    my $lineno        = 0;
 
     for my $f (@$file) {
         $fileno++;
@@ -251,13 +332,13 @@ sub lines {
             $lineno++;
             chomp;
 
-            if ( ( $fileno == 1 ) && ( $lineno == 1 ) && defined( $self->{header_handler} ) ) {
-                $self->{header_handler}($_);
-            } elsif ( ( $fileno == 1 ) && ( $lineno == 1 ) && $self->{header_skip} ) {
-                # Do nothing;
-            } else {
-                push @lines, $_;
+            if ( $lineno == 1 ) {
+                if ( $self->_handle_header( $f, $_, 0, $fileno - 1 ) ) {
+                    next;
+                }
             }
+
+            push @lines, $_;
         }
 
         close $fh;
@@ -266,15 +347,19 @@ sub lines {
     return @lines;
 }
 
-# Internal function to read header line
+# Internal function to read header line (if we need to)
 sub _read_header {
-    my ( $self, $file ) = @_;
+    my ( $self, $file, $fileno ) = @_;
 
     my ( $fh, undef ) = _open_and_seek( $file, 1, 0 );
     my $line = <$fh>;
     close $fh;
 
-    chomp($line) if defined $line;
+    if ( defined($line) ) {
+        chomp($line);
+        $self->_handle_header( $file, $line, 0, $fileno );
+    }
+
     return $line;
 }
 
@@ -288,12 +373,15 @@ sub _read_header {
 sub _forlines_chunk {
     my ( $self, $code, $file, $part ) = @_;
 
-    my $fileno = 0;
-    my $lineno = 0;
+    my $fileno        = 0;
+    my $lineno        = 0;
+    my $extended_info = $self->{extended_info};
 
     for my $f (@$file) {
         $fileno++;
 
+        my $extended = $self->_extended( $f, $part );
+
         my $procs = $self->{processes};
         my ( $fh, $end ) = _open_and_seek( $f, $procs, $part );
 
@@ -302,21 +390,14 @@ sub _forlines_chunk {
 
             chomp;
 
-            # Handle header option
-            if (   ( !$part )
-                && ( $fileno == 1 )
-                && ( $lineno == 1 )
-                && ( defined( $self->{header_handler} ) ) )
-            {
-                # Do nothing, we're skipping the header.
-            } elsif ( ( !$part )
-                && ( $fileno == 1 )
-                && ( $lineno == 1 )
-                && ( $self->{header_skip} ) )
-            {
-                # Do nothing, we're skipping the header.
+            if ( $lineno == 1 && $self->_handle_header( $f, $_, $part, $fileno - 1 ) ) {
+                # Do nothing, we handled the header.
             } else {
-                $code->($_);
+                if ($extended_info) {
+                    $code->( $_, $extended );
+                } else {
+                    $code->($_);
+                }
             }
 
             # If we're reading multi-parts, do we need to end the read?
@@ -329,36 +410,39 @@ sub _forlines_chunk {
     return $lineno;
 }
 
-# Internal function to perform a grep on a single chunk of the file.
+# Internal function to perform a map/grep on a single chunk of the file.
 #
 # Procs should be >= 1.  It represents the number of chunks the file
 # has.
 #
 # Part should be >= 0 and < Procs.  It represents the zero-indexed chunk
 # number this invocation is processing.
-sub _grep_chunk {
-    my ( $self, $code, $file, $procs, $part ) = @_;
+#
+# isgrep = true if we want to just apply the code as a grep, not as a
+# map.
+sub _grepmap_chunk {
+    my ( $self, $code, $file, $isgrep, $procs, $part ) = @_;
 
-    my @lines;
-    my $fileno = 0;
-    my $lineno = 0;
+    my @mapped_lines;
+    my $fileno        = 0;
+    my $lineno        = 0;
+    my $extended_info = $self->{extended_info};
 
     for my $f (@$file) {
         $fileno++;
 
+        my $extended = $self->_extended( $f, $part );
+
         my ( $fh, $end ) = _open_and_seek( $f, $procs, $part );
 
+        my @filelines;
         while (<$fh>) {
             $lineno++;
 
             chomp;
 
-            if (   ( !$part )
-                && ( $fileno == 1 )
-                && ( $lineno == 1 )
-                && ( defined( $self->{header_handler} ) ) )
-            {
-                $self->{header_handler}($_);
+            if ( $lineno == 1 && $self->_handle_header( $f, $_, $part, $fileno - 1 ) ) {
+                # Do nothing, we handled the header.
             } elsif ( ( !$part )
                 && ( $fileno == 1 )
                 && ( $lineno == 1 )
@@ -366,64 +450,30 @@ sub _grep_chunk {
             {
                 # Do nothing, we're skipping the header.
             } else {
-                if ( $code->($_) ) {
-                    push @lines, $_;
+                if ($isgrep) {
+                    if ($extended_info) {
+                        if ( $code->( $_, $extended ) ) {
+                            push @filelines, $_;
+                        }
+                    } else {
+                        if ( $code->($_) ) {
+                            push @filelines, $_;
+                        }
+                    }
+                } else {
+                    # We are doing a map
+                    if ($extended_info) {
+                        push @filelines, $code->( $_, $extended );
+                    } else {
+                        push @filelines, $code->($_);
+                    }
                 }
             }
 
             # If we're reading multi-parts, do we need to end the read?
             if ( ( $end > 0 ) && ( tell($fh) > $end ) ) { last; }
         }
-
-        close $fh;
-    }
-
-    return \@lines;
-}
-
-# Internal function to perform a map on a single chunk of the file.
-#
-# Procs should be >= 1.  It represents the number of chunks the file
-# has.
-#
-# Part should be >= 0 and < Procs.  It represents the zero-indexed chunk
-# number this invocation is processing.
-sub _map_chunk {
-    my ( $self, $code, $file, $procs, $part ) = @_;
-
-    my @mapped_lines;
-    my $fileno = 0;
-    my $lineno = 0;
-
-    for my $f (@$file) {
-        $fileno++;
-
-        my ( $fh, $end ) = _open_and_seek( $f, $procs, $part );
-
-        while (<$fh>) {
-            $lineno++;
-
-            chomp;
-
-            if (   ( !$part )
-                && ( $fileno == 1 )
-                && ( $lineno == 1 )
-                && ( defined( $self->{header_handler} ) ) )
-            {
-                $self->{header_handler}($_);
-            } elsif ( ( !$part )
-                && ( $fileno == 1 )
-                && ( $lineno == 1 )
-                && ( $self->{header_skip} ) )
-            {
-                # Do nothing, we're skipping the header.
-            } else {
-                push @mapped_lines, $code->($_);
-            }
-
-            # If we're reading multi-parts, do we need to end the read?
-            if ( ( $end > 0 ) && ( tell($fh) > $end ) ) { last; }
-        }
+        push @mapped_lines, \@filelines;
 
         close $fh;
     }
@@ -553,11 +603,95 @@ sub _listlike {
     my $thing = shift;
 
     if ( reftype($thing) ) { return 1; }
-    if ( defined(blessed($thing)) && overload::Method( $thing, '[]' ) ) { return 1; }
+    if ( defined( blessed($thing) ) && overload::Method( $thing, '[]' ) ) { return 1; }
 
     return;
 }
 
+# Takes a hashref, key, and default value
+# If the hashref item exists, returns the corresponding value.  If the hashref
+# item does not exist, returns the default value.
+sub _option_helper {
+    if ( scalar(@_) != 3 ) { confess 'invalid call' }
+    my ( $hash, $key, $default ) = @_;
+
+    if ( exists( $hash->{$key} ) ) {
+        return $hash->{$key};
+    } else {
+        return $default;
+    }
+}
+
+sub _is_number {
+    if ( scalar(@_) != 1 ) { confess 'invalid call' }
+    my $val = shift;
+
+    if ( !defined($val) ) { return; }
+
+    return $val =~ /
+            \A              # Start of string
+            [0-9]+          # ASCII digit
+            (?: \. 0+)?     # Optional .0 or .000 or .00000 etc
+            \z              # End of string
+        /sxx;
+}
+
+# Returns an extended info object
+sub _extended {
+    if ( scalar(@_) != 3 ) { confess 'invalid call' }
+    my ( $self, $filename, $process_number ) = @_;
+
+    return {
+        filename       => $filename,
+        object         => $self,
+        process_number => $process_number,
+    };
+}
+
+# Executes the header_handler function when required, or skipps headers.
+#
+# This returns TRUE if there is a header to process.  FALSE otherwise
+#
+# This takes several parameters:
+#   $self - This is an object method of course.
+#   $filename = The filename being processed
+#   $line - The line to process
+#   $part - Which "part" is calling this (we always return FALSE and
+#           refuse to process the header if $part > 0)
+#   $fileno - Which file number are we on (start at zero)
+#
+# If header_skip is FALSE and header_handler is unset, this ALWAYS
+# returns false.
+#
+# This should never be called except for the first line of a file
+sub _handle_header {
+    if ( scalar(@_) != 5 ) { confess 'invalid call' }
+    my ( $self, $filename, $line, $part, $fileno ) = @_;
+
+    if ($part) { return; }
+
+    if ( ( !$self->header_skip() ) && ( !defined( $self->header_handler() ) ) ) {
+        return;
+    }
+
+    if ( $fileno && ( !$self->header_all_files() ) ) {
+        return;
+    }
+
+    # We have a header to process.
+    if ( defined( $self->header_handler() ) ) {
+        local $_ = $line;
+
+        if ( $self->{extended_info} ) {
+            my $extended = $self->_extended( $filename, $part );
+            $self->{header_handler}( $line, $extended );
+        } else {
+            $self->{header_handler}($line);
+        }
+    }
+    return 1;
+}
+
 
 1;
 
@@ -573,7 +707,7 @@ File::ByLine::Object - Internal object used by File::ByLine
 
 =head1 VERSION
 
-version 1.181950
+version 1.181951
 
 =head1 SEE File::ByLine
 
